from __future__ import annotations

import sys
from argparse import ArgumentParser, Namespace
from typing import Optional, TYPE_CHECKING, Protocol, Any, Union

import pkg_resources


# Circumvent mypy/pylint shenanigans ~
class _SubParsersAction:  # pylint: disable= too-few-public-methods # typechecker only, ignore warnings
    def add_parser(  # pylint: disable=redefined-builtin, unused-argument # typechecker only, ignore warnings
        self,
        name: str,
        *,
        prog: Optional[str] = None,
        aliases: Optional[Any] = None,
        help: Optional[str] = None,
        **kwargs: Any,
    ) -> ArgumentParser:
        raise NotImplementedError


class CliEntrypoint(Protocol):  # pylint: disable= too-few-public-methods
    def __call__(self, parent: Optional[_SubParsersAction]) -> Optional[int]:
        raise NotImplementedError


class _CliPlugin:  # pylint: disable= too-few-public-methods
    def __init__(self, parser: ArgumentParser):
        self.parser = parser

    def _run(self, ns: Namespace) -> int:
        if not hasattr(ns, "cmd"):
            raise NotImplementedError(
                "Command defined in argparse, but it's function was not specified."
            )
        cmd = ns.cmd
        result: Optional[int] = cmd(ns)
        if result is None:  # Assume success
            result = 0
        return result

    def run_with(self, *args: Any) -> Union[str, int, None]:
        try:
            ns = self.parser.parse_args(args)
            return self._run(ns)
        except SystemExit as sys_exit:
            return sys_exit.code

    def run(self) -> None:
        ns = self.parser.parse_args()
        exit_code = self._run(ns)
        sys.exit(exit_code)


class CliPluginGroup(_CliPlugin):  # pylint: disable= too-few-public-methods
    GROUP: str = None  # type: ignore

    def __init__(
        self,
        parent: Optional[_SubParsersAction] = None,
    ):
        if TYPE_CHECKING:
            self.subparsers = None
        if self.GROUP is None:
            raise ValueError
        parser = self._create_parser(parent)
        super().__init__(parser)
        self.subparsers = self._create_subparser_group(parser)
        self._load()

    def _create_parser(
        self, command_group: Optional[_SubParsersAction] = None
    ) -> ArgumentParser:
        raise NotImplementedError

    def _create_subparser_group(self, parser: ArgumentParser) -> _SubParsersAction:
        return parser.add_subparsers()  # type: ignore

    def _load(self) -> None:
        for ep in pkg_resources.iter_entry_points(group=self.GROUP):
            ep_func: CliEntrypoint = ep.load()
            ep_func(parent=self.subparsers)


class CliPlugin(_CliPlugin):  # pylint: disable= too-few-public-methods
    def __init__(self, parent: Optional[_SubParsersAction] = None):
        parser = self._create_parser(parent)
        super().__init__(parser)
        if self.parser.get_default("cmd") is None:
            self.parser.set_defaults(cmd=self.command)

    def _create_parser(
        self, command_group: Optional[_SubParsersAction] = None
    ) -> ArgumentParser:
        raise NotImplementedError

    def command(self, ns: Namespace) -> Optional[int]:
        raise NotImplementedError


class RelicCli(CliPluginGroup):  # pylint: disable= too-few-public-methods
    GROUP = "relic.cli"

    def _create_parser(
        self, command_group: Optional[_SubParsersAction] = None
    ) -> ArgumentParser:
        if command_group is None:
            return ArgumentParser("relic")
        return command_group.add_parser("relic")


cli_root = RelicCli()



import subprocess

# Local testing requires running `pip install -e "."`
from contextlib import redirect_stdout
import io
from typing import Sequence

import pytest


class CommandTests:
    def test_run(self, args: Sequence[str], output: str, exit_code: int):
        _args = ["relic", *args]
        cmd = subprocess.run(_args, capture_output=True, text=True)
        result = cmd.stdout
        status = cmd.returncode
        print(f"'{result}'")  # Visual Aid for Debugging
        assert output in result
        assert status == exit_code

    def test_run_with(self, args: Sequence[str], output: str, exit_code: int):
        from relic.core.cli import cli_root
        with io.StringIO() as f:
            with redirect_stdout(f):
                status = cli_root.run_with(*args)
            f.seek(0)
            result = f.read()
            print(f"'{result}'")  # Visual Aid for Debugging
            assert output in result
            assert status == exit_code

_HELP = ["-h"], """usage: relic [-h] {} ...""", 0

_TESTS = [_HELP]
_TEST_IDS = [' '.join(_[0]) for _ in _TESTS]


@pytest.mark.parametrize(["args", "output", "exit_code"], _TESTS, ids=_TEST_IDS)
class TestRelicCli(CommandTests):
    ...



"""
Tests which ensures releases do not break backwards-compatibility by failing to expose modules/names
"""

import importlib
from typing import List, Iterable, Tuple

import pytest

core__all__ = [
    "errors",
    "typeshed",
]


@pytest.mark.parametrize("submodule", core__all__)
def test_import_module(submodule: str):
    try:
        importlib.import_module(f"relic.core.{submodule}")
    except ImportError:
        raise AssertionError(f"{submodule} is no longer exposed!")


errors__all__ = [
    "MismatchError",
    "RelicToolError",
    "T",
]
typeshed__all__ = [
    "TypeAlias",
]


def module_imports_helper(submodule: str, all: List[str]) -> Iterable[Tuple[str, str]]:
    return zip([submodule] * len(all), all)


@pytest.mark.parametrize(
    ["submodule", "attribute"],
    [
        *module_imports_helper("errors", errors__all__),
        *module_imports_helper("typeshed", typeshed__all__),
    ],
)
def test_module_imports(submodule: str, attribute: str):
    module = importlib.import_module(f"relic.core.{submodule}")
    _ = getattr(module, attribute)
